@@ -1,11 +1,11 @@
 (* approx: proxy server for Debian archive files
-   Copyright (C) 2010  Eric C. Cooper <ecc@cmu.edu>
+   Copyright (C) 2011  Eric C. Cooper <ecc@cmu.edu>
    Released under the GNU General Public License *)
 
 open Config_file
 open Util
 
-let version = "4.5"
+let version = "5.0"
 
 let default_config = "/etc/approx/approx.conf"
 
@@ -40,7 +40,7 @@ let split_cache_path path =
     let j = String.index_from path i '/' in
     substring path ~from: i ~until: j, substring path ~from: (j + 1)
   else
-    invalid_arg "split_cache_path"
+    invalid_string_arg "split_cache_path" path
 
 let shorten path =
   if is_prefix cache_dir path then
@@ -48,6 +48,14 @@ let shorten path =
   else
     path
 
+let check_current_directory () =
+  let cwd = Sys.getcwd () in
+  if cwd <> cache_dir then
+    failwith (Printf.sprintf "current directory is %s, not %s" cwd cache_dir)
+
+let interval = get_int "$interval" ~default: 720
+let params = ("$interval", string_of_int interval) :: params
+
 let max_rate = get "$max_rate" ~default: "unlimited"
 let params = ("$max_rate", max_rate) :: params
 
@@ -89,7 +97,7 @@ let repository_table items =
   String.concat ""
     (List.map
        (fun (k, v) ->
-          "<tr><td>" ^ k ^ "</td>\
+          "<tr><td><a href=\"/" ^ k ^ "\">" ^ k ^ "</a></td>\
                <td><a href=\"" ^ v ^ "\">" ^ v ^ "</a></td></tr>\n")
        (sort_config items))
 
